@@ -37,9 +37,17 @@
 #include <libxml/parser.h>
 #include <time.h>
 
+#define DEFAULT_NOTIFICATION_SOUND_PATH		SOUNDS_DIR "/clicked.wav"
+
 #define TIMER_APPLET_DATA_PATH			"/.gnome2/timer-applet"
 #define TIMER_APPLET_DATA_FILE_VERSION		"1.0"
 #define TIMER_APPLET_MENU_FILE_NAME		"GNOME_TimerApplet.xml"
+#define PRESETS_SUBMENU_PATH			"/popups/popup/Presets"
+#define TA_MENU_COMMAND_PAUSE_TIMER		"/commands/PauseTimer"
+#define TA_MENU_COMMAND_RESUME_TIMER		"/commands/ResumeTimer"
+#define TA_MENU_COMMAND_RESTART_TIMER		"/commands/RestartTimer"
+#define TA_MENU_COMMAND_STOP_TIMER		"/commands/StopTimer"
+#define TA_DONT_SHOW_ZEROS			TRUE
 
 #define TIMER_TIMEOUT_INTERVAL			500
 
@@ -48,6 +56,16 @@
  * Callback prototypes
  */
 static void
+on_pause_timer_activate (BonoboUIComponent *component, TimerApplet *timer_applet, const gchar *cname);
+static void
+on_resume_timer_activate (BonoboUIComponent *component, TimerApplet *timer_applet, const gchar *cname);
+static void
+on_restart_timer_activate (BonoboUIComponent *component, TimerApplet *timer_applet, const gchar *cname);
+static void
+on_stop_timer_activate (BonoboUIComponent *component, TimerApplet *timer_applet, const gchar *cname);
+static void
+on_presets_submenu_item_activate (BonoboUIComponent *uic, TimerApplet *timer_applet, const gchar *verb);
+static void
 on_manage_presets_activate (BonoboUIComponent *component, TimerApplet *timer_applet, const gchar *cname);
 static void
 on_preferences_activate (BonoboUIComponent *component, TimerApplet *timer_applet, const gchar *cname);
@@ -64,6 +82,8 @@ on_preferences_change (GConfClient *client, guint cnxn_id, GConfEntry *entry, Ti
 static gboolean
 on_timer_timeout (TimerApplet *timer_applet);
 static void
+on_presets_changed (GObject *obj, TimerApplet *timer_applet);
+static void
 on_applet_change_background (PanelApplet *applet, PanelAppletBackgroundType type,
 				GdkColor *color, GdkPixmap *pixmap, gpointer user_data);
 static void
@@ -75,6 +95,10 @@ on_applet_instance_destroy (GtkWidget *applet, TimerApplet *timer_applet);
 
 
 static const BonoboUIVerb timer_applet_verbs [] = {
+	BONOBO_UI_UNSAFE_VERB ("PauseTimer", on_pause_timer_activate),
+	BONOBO_UI_UNSAFE_VERB ("ResumeTimer", on_resume_timer_activate),
+	BONOBO_UI_UNSAFE_VERB ("RestartTimer", on_restart_timer_activate),
+	BONOBO_UI_UNSAFE_VERB ("StopTimer", on_stop_timer_activate),
 	BONOBO_UI_UNSAFE_VERB ("ManagePresets", on_manage_presets_activate),
 	BONOBO_UI_UNSAFE_VERB ("Preferences", on_preferences_activate),
 	BONOBO_UI_UNSAFE_VERB ("Help", on_help_activate),
@@ -83,6 +107,7 @@ static const BonoboUIVerb timer_applet_verbs [] = {
 };
 
 static GtkListStore *timer_presets_list = NULL;
+static guint presets_changed_sig_id = 0;
 static GtkWidget *manage_presets_window = NULL;
 static GtkWidget *about_dialog;
 
@@ -132,48 +157,57 @@ set_tooltip_to_stop_str (TimerApplet *timer_applet)
  * the remaining amount of time. The caller should free the returned string
  * when it's finished being used.
  * show_all indicates whether this function should return hours:minutes:seconds.
+ * shorten_zero indicates whether a "shortened" string (currenty, an empty string)
+ * should be returned when remaining_seconds is equal to zero.
  * If show_all == false, it returns either hours:minutes or minutes:seconds,
  * depending on how much time is remaining, so that the user isn't given
  * information he doesn't need.
  */
 static gchar *
-construct_time_str (gdouble remaining_seconds, gboolean show_all)
+construct_time_str (gdouble remaining_seconds, gboolean show_all, gboolean shorten_zero)
 {
 	guint hours, minutes, seconds;
 	gchar *time_str;
 
 	g_assert (remaining_seconds >= 0);
 	
-	hours = (guint)(remaining_seconds / 3600);
-	minutes = (guint)((remaining_seconds - (hours * 3600)) / 60);
-	seconds = (guint)(remaining_seconds - (hours *3600) - (minutes * 60));
+	if (!shorten_zero || remaining_seconds > 0)
+	{
+		hours = (guint)(remaining_seconds / 3600);
+		minutes = (guint)((remaining_seconds - (hours * 3600)) / 60);
+		seconds = (guint)(remaining_seconds - (hours *3600) - (minutes * 60));
 
-	g_assert (hours >= 0);
-	g_assert (minutes >= 0);
-	g_assert (seconds >= 0);
-	g_assert (hours <= MAX_NUM_HOURS);
-	g_assert (minutes <= 59);
-	g_assert (seconds <= 59);
+		g_assert (hours >= 0);
+		g_assert (minutes >= 0);
+		g_assert (seconds >= 0);
+		g_assert (hours <= MAX_NUM_HOURS);
+		g_assert (minutes <= 59);
+		g_assert (seconds <= 59);
 	
-	/* strip hour indication field if we don't time hours and skip seconds
-	 * indicator if we are.
-	 */
+		/* strip hour indication field if we don't time hours and skip seconds
+		 * indicator if we are.
+		 */
 
-	if (show_all) {
-		/* hours:minutes:seconds */
-		time_str = g_strdup_printf (_("%.2d:%.2d:%.2d"), hours, minutes, seconds);
-	}
-	else {
-		if ((hours > 0) || (minutes > 14)) {
-			/* hours:minutes */
-			time_str = g_strdup_printf (_("%.2d:%.2d"), hours, minutes);
+		if (show_all) {
+			/* hours:minutes:seconds */
+			time_str = g_strdup_printf (_("%.2d:%.2d:%.2d"), hours, minutes, seconds);
 		}
 		else {
-			/* minutes:seconds */
-			time_str = g_strdup_printf (_("%.2d:%.2d"), minutes, seconds);
+			if ((hours > 0) || (minutes > 14)) {
+				/* hours:minutes */
+				time_str = g_strdup_printf (_("%.2d:%.2d"), hours, minutes);
+			}
+			else {
+				/* minutes:seconds */
+				time_str = g_strdup_printf (_("%.2d:%.2d"), minutes, seconds);
+			}
 		}
 	}
-	
+	else
+	{
+		time_str = g_strdup ("");
+	}
+
 	return time_str;
 }
 
@@ -188,7 +222,7 @@ construct_display_name (const gchar *name, gdouble duration)
 	gchar *display_name;
 	gchar *time_str;
 	
-	time_str = construct_time_str (duration, TRUE /* always show hours:minutes:seconds */);
+	time_str = construct_time_str (duration, TRUE /* always show hours:minutes:seconds */, FALSE /* don't shorten zero */);
 	/* <name of preset> (<number of hours, minutes, and seconds>) */
 	display_name = g_strdup_printf (_("%s (%s)"), name, time_str);
 	g_free (time_str);
@@ -269,8 +303,7 @@ construct_applet_content (TimerApplet *timer_applet)
 		/* calc_remaining_time may return a negative number if the user decides to show
 		 * the remaining time just as the timer finishes its countdown
 		 */
-		remaining_time_str = construct_time_str (MAX (0, calc_remaining_time (timer_applet)), FALSE);
-		
+		remaining_time_str = construct_time_str (MAX (0, calc_remaining_time (timer_applet)), FALSE, TA_DONT_SHOW_ZEROS);
 		timer_applet->main_remaining_time_label = gtk_label_new (remaining_time_str);
 		g_free (remaining_time_str);
 		remaining_time_str = NULL;
@@ -385,6 +418,8 @@ save_timer_presets (void)
 	g_assert (doc);
 	xmlFreeDoc (doc);
 	
+	g_print ("Saved presets.\n");
+	g_signal_emit (timer_presets_list, presets_changed_sig_id, 0, NULL);
 }
 
 /**
@@ -466,7 +501,8 @@ set_ui_time (TimerApplet *timer_applet, gdouble seconds_remaining)
 
 	if (timer_applet->main_remaining_time_label) {
 		gchar *applet_time_str;
-		applet_time_str = construct_time_str (seconds_remaining, FALSE);
+		applet_time_str = construct_time_str (seconds_remaining, FALSE, TA_DONT_SHOW_ZEROS);
+		g_assert (GTK_IS_LABEL (timer_applet->main_remaining_time_label));
 		gtk_label_set_text (GTK_LABEL (timer_applet->main_remaining_time_label), applet_time_str);
 		g_free (applet_time_str);
 		applet_time_str = NULL;
@@ -477,9 +513,10 @@ set_ui_time (TimerApplet *timer_applet, gdouble seconds_remaining)
 		gchar *tooltip_time_str;
 		
 		/* Always show hours:minutes:seconds in tooltip */
-		tooltip_time_str = construct_time_str (seconds_remaining, TRUE);
+		tooltip_time_str = construct_time_str (seconds_remaining, TRUE, FALSE /* don't shorten zero */);
 		
-		if (timer_applet->cur_profile_name) {
+		g_assert (timer_applet->cur_profile_name);
+		if (strlen (timer_applet->cur_profile_name) > 0) {
 			/* <remaining time in hours:minutes:seconds> (<name of preset>) */
 			tooltip_str = g_strdup_printf (_("%s (%s)"), tooltip_time_str, timer_applet->cur_profile_name);
 		}
@@ -526,9 +563,37 @@ timer_is_paused (TimerApplet *timer_applet)
 static void
 activate_timer (TimerApplet *timer_applet)
 {
+	BonoboUIComponent *popup_component;
+
 	g_assert (timer_applet);
 	g_assert (timer_applet->timer);
 	g_assert (timer_applet->timer_running);
+	
+	popup_component = panel_applet_get_popup_component (PANEL_APPLET (timer_applet->applet));
+
+	/* Show the PauseTimer menu item. */
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_PAUSE_TIMER,
+					"hidden", "0",
+					NULL);
+
+	/* Hide the ResumeTimer menu item. */
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_RESUME_TIMER,
+					"hidden", "1",
+					NULL);
+
+	/* Show the RestartTimer menu item. */
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_RESTART_TIMER,
+					"hidden", "0",
+					NULL);
+
+	/* Show the StopTimer menu item. */
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_STOP_TIMER,
+					"hidden", "0",
+					NULL);
 
 	g_assert (timer_applet->main_button_layout_box);
 	{
@@ -553,20 +618,20 @@ resume_timer (TimerApplet *timer_applet)
 
 	g_timer_continue (timer_applet->timer);
 	timer_applet->timer_running = TRUE;
-	
+		
 	activate_timer (timer_applet);
 }
 
 /**
  * Pause the timer. Note: the user may pause the timer just as
  * the timer is about to finish, meaning that calc_remaining_time
- * may return a negative value. Also, calc_remaining_time will always
- * return a negative value when pause_timer is called from stop_timer,
- * because stop_timer is only called when calc_remaining_time is negative.
+ * may return a negative value.
  */
 static void
 pause_timer (TimerApplet *timer_applet)
 {
+	BonoboUIComponent *popup_component;
+
 	g_assert (timer_applet != NULL);
 	g_assert (timer_applet->timer);
 	
@@ -574,32 +639,76 @@ pause_timer (TimerApplet *timer_applet)
 	timer_applet->timer_running = FALSE;
 
 	set_ui_time (timer_applet, MAX (0, calc_remaining_time (timer_applet)));
-	set_tooltip_to_stop_str (timer_applet);
+	set_tooltip (timer_applet, _("Paused. Click to resume timer."));
 	g_assert (timer_applet->main_button_layout_box);
 	g_object_set (G_OBJECT (timer_applet->main_button_layout_box), "sensitive", FALSE, NULL);
+
+	popup_component = panel_applet_get_popup_component (PANEL_APPLET (timer_applet->applet));
+
+	/* Hide the PauseTimer menu item. */
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_PAUSE_TIMER,
+					"hidden", "1",
+					NULL);
+
+	/* Show the ResumeTimer menu item. */
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_RESUME_TIMER,
+					"hidden", "0",
+					NULL);
 }
 
 /**
- * Stop the timer. Should only be called when the timer has finished a countdown.
+ * Stop the timer. Can be called when the timer has finished a countdown,
+ * or when the user explicitly resets the timer to zero.
  */
 static void
 stop_timer (TimerApplet *timer_applet)
 {
-	/* Must be a negative number because on_timer_timeout only
-	 * calls stop_timer if remaining time is a negative number
-	 */
-	g_assert (calc_remaining_time (timer_applet) < 0);
+	BonoboUIComponent *popup_component;
 
-	/* Since we know that the remaining time is a negative number,
-	 * we know that pause_timer will automatically set the applet's
-	 * remaining time label to 00:00, so we don't need do anything
-	 * to the label in this function.
+	g_assert (timer_applet);
+	g_assert (timer_applet->timer);
+	g_assert (timer_is_running (timer_applet) || timer_is_paused (timer_applet));
+
+	/* calc_remaining_time (timer_applet) can be a negative number
+	 * if called from on_timer_timeout when the timer finishes,
+	 * or it can be >= 0 if the user explicitly resets the timer to zero
+	 * while the timer is running or paused.
 	 */
-	pause_timer (timer_applet);
+
+	g_timer_stop (timer_applet->timer);
 	g_timer_reset (timer_applet->timer);
 	g_timer_destroy (timer_applet->timer);
 	timer_applet->timer = NULL;
+	timer_applet->timer_running = FALSE;
 	g_assert (calc_remaining_time (timer_applet) == 0);
+
+	set_ui_time (timer_applet, 0);
+	set_tooltip_to_stop_str (timer_applet);
+
+	g_assert (timer_applet->main_button_layout_box);
+	g_object_set (G_OBJECT (timer_applet->main_button_layout_box), "sensitive", FALSE, NULL);
+
+	popup_component = panel_applet_get_popup_component (PANEL_APPLET (timer_applet->applet));
+
+	/* Hide the PauseTimer menu item. */
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_PAUSE_TIMER,
+					"hidden", "1",
+					NULL);
+
+	/* Hide the RestartTimer menu item. */
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_RESTART_TIMER,
+					"hidden", "1",
+					NULL);
+
+	/* Hide the StopTimer menu item. */
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_STOP_TIMER,
+					"hidden", "1",
+					NULL);
 }
 
 /**
@@ -645,7 +754,9 @@ notify_timer_done (TimerApplet *timer_applet)
 	g_assert (timer_applet);
 
 	cur_time_str = construct_current_time_str ();
-	if (timer_applet->cur_profile_name) {
+
+	g_assert (timer_applet->cur_profile_name);
+	if (strlen (timer_applet->cur_profile_name) > 0) {
 		/* "<timer name>" finished at <finish time> */
 		timer_done_str = g_strdup_printf (_("\"%s\" finished at %s"),
 							timer_applet->cur_profile_name,
@@ -661,20 +772,44 @@ notify_timer_done (TimerApplet *timer_applet)
 	dialog = gtk_message_dialog_new (NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
 					GTK_MESSAGE_INFO, GTK_BUTTONS_OK, timer_done_str);
 	g_free (timer_done_str);
-	g_object_set (G_OBJECT (dialog), "title", _("Timer Finished"), NULL);
 	timer_done_str = NULL;
+	g_object_set (G_OBJECT (dialog), "title", _("Timer Finished"), NULL);
 	
 	if (panel_applet_gconf_get_bool (timer_applet->applet, PLAY_NOTIFICATION_SOUND_KEY, NULL)) {
-		g_assert (timer_applet->notification_sound_path);
-		g_print ("Playing notification sound: %s\n", timer_applet->notification_sound_path);
-		gnome_sound_play (timer_applet->notification_sound_path);
+		gboolean use_custom_sound;
+		gchar *notification_sound_path;
+		use_custom_sound = panel_applet_gconf_get_bool (timer_applet->applet, USE_CUST_NOTIFICATION_SOUND_KEY, NULL);
+
+		if (use_custom_sound) {
+			g_print ("Using custom notification sound.\n");
+			notification_sound_path = panel_applet_gconf_get_string (timer_applet->applet,
+										CUST_NOTIFICATION_SOUND_PATH_KEY,
+										NULL);
+			g_assert (notification_sound_path);
+			if (strlen (notification_sound_path) == 0) {
+				g_print ("No custom notification sound specified. Using default sound.\n");
+				g_free (notification_sound_path);
+				notification_sound_path = g_strdup (DEFAULT_NOTIFICATION_SOUND_PATH);
+			}
+		}
+		else {
+			g_print ("Using default notification sound.\n");
+			notification_sound_path = g_strdup (DEFAULT_NOTIFICATION_SOUND_PATH);
+		}
+		
+		g_assert (notification_sound_path);
+		g_print ("Playing notification sound: %s\n", notification_sound_path);
+		gnome_sound_play (notification_sound_path);
+
+		g_free (notification_sound_path);
+		notification_sound_path = NULL;
 	}
 
 	/* Make the applet button un-clickable until the user dismisses the dialog box */
 
 	g_assert (timer_applet->main_button);
 	g_object_set (G_OBJECT (timer_applet->main_button), "sensitive", FALSE, NULL);
-	
+
 	if (panel_applet_gconf_get_bool (timer_applet->applet, NOTIFICATION_DIALOG_ALWAYS_ON_TOP_KEY, NULL)) {
 		gtk_window_set_keep_above (GTK_WINDOW (dialog), TRUE);
 	}
@@ -682,8 +817,17 @@ notify_timer_done (TimerApplet *timer_applet)
 	if (panel_applet_gconf_get_bool (timer_applet->applet, NOTIFICATION_DIALOG_STICKY_KEY, NULL)) {
 		gtk_window_stick (GTK_WINDOW (dialog));
 	}
+
+	if (timer_applet->main_remaining_time_label) {
+		gtk_label_set_text (GTK_LABEL (timer_applet->main_remaining_time_label), _("Finished"));
+	}
 	
 	gtk_dialog_run (GTK_DIALOG (dialog));
+
+	if (timer_applet->main_remaining_time_label) {
+		gtk_label_set_text (GTK_LABEL (timer_applet->main_remaining_time_label), "");
+	}
+
 	g_object_set (G_OBJECT (timer_applet->main_button), "sensitive", TRUE, NULL);
 	gtk_widget_destroy (dialog);
 }
@@ -724,6 +868,80 @@ row_ref_from_iter (GtkTreeModel *tree_model, GtkTreeIter *tree_iter)
 	return new_row_ref;
 }
 
+static void 
+update_presets_submenu (TimerApplet *timer_applet)
+{
+	BonoboUIComponent *popup_component;
+
+	popup_component = panel_applet_get_popup_component (PANEL_APPLET (timer_applet->applet));
+	
+	/* Remove existing menu items. */
+	int i;
+	for (i = 0;; i++) {
+		gchar *item_path = g_strdup_printf (PRESETS_SUBMENU_PATH "/Preset_%d", i);
+		if (bonobo_ui_component_path_exists (popup_component, item_path, NULL)) {
+			bonobo_ui_component_rm (popup_component, item_path, NULL);
+			g_print ("Removed presets submenu item: %s\n", item_path);
+			g_free (item_path);
+			item_path = NULL;
+		}
+		else {
+			g_free (item_path);
+			item_path = NULL;
+			break;
+		}
+	}
+
+	{
+		GtkTreeIter iter;
+		gboolean valid;
+		
+		valid = gtk_tree_model_get_iter_first (GTK_TREE_MODEL (timer_presets_list), &iter);
+		while (valid) {
+			gchar *preset_name;
+			gdouble preset_duration;
+			gchar *preset_display_name;
+			gint preset_index;
+			
+			gtk_tree_model_get (GTK_TREE_MODEL (timer_presets_list), &iter,
+								PROFILE_NAME_COL, &preset_name,
+								PROFILE_DURATION_COL, &preset_duration,
+								-1);
+			g_assert (preset_name);
+			preset_display_name = construct_display_name (preset_name, preset_duration);
+			g_free (preset_name);
+			preset_name = NULL;
+			
+			/* Determine this preset's index. */
+			{
+				GtkTreeRowReference *row_ref = row_ref_from_iter (GTK_TREE_MODEL (timer_presets_list), &iter);
+				GtkTreePath *path = gtk_tree_row_reference_get_path (row_ref);
+				gint *path_indices = gtk_tree_path_get_indices (path);
+				g_assert (gtk_tree_path_get_depth (path) == 1);
+				preset_index = path_indices[0];
+				gtk_tree_path_free (path);
+				path = NULL;
+			}
+			
+			gchar *verb = g_strdup_printf ("Preset_%d", preset_index);
+			BonoboUINode *node = bonobo_ui_node_new ("menuitem");	
+			bonobo_ui_node_set_attr (node, "name", verb);
+			bonobo_ui_node_set_attr (node, "verb", verb);
+			bonobo_ui_node_set_attr (node, "label", preset_display_name);
+			g_print ("Added presets submenu item: %s\n", verb);
+			bonobo_ui_component_set_tree (popup_component, PRESETS_SUBMENU_PATH, node, NULL);
+			bonobo_ui_component_add_verb (popup_component, verb, (BonoboUIVerbFn)on_presets_submenu_item_activate, timer_applet);
+			g_free (verb);
+			verb = NULL;
+			
+			g_free (preset_display_name);
+			preset_display_name = NULL;
+
+			valid = gtk_tree_model_iter_next (GTK_TREE_MODEL (timer_presets_list), &iter);
+		}
+	}
+}
+
 /**
  * Shows a dialog box to ask whether the user would like to
  * resume a paused timer. This function resumes the timer or
@@ -760,8 +978,10 @@ ask_user_resume_timer (TimerApplet *timer_applet)
 		main_vbox = GTK_DIALOG (ask_resume_dialog)->vbox;
 		hbox = gtk_hbox_new (FALSE, 0);
 		image = gtk_image_new_from_stock (GTK_STOCK_DIALOG_QUESTION, GTK_ICON_SIZE_DIALOG);
-		remaining_time_str = construct_time_str (timer_applet_get_remaining_time (timer_applet), TRUE);
-		if (timer_applet->cur_profile_name) {
+		remaining_time_str = construct_time_str (timer_applet_get_remaining_time (timer_applet), TRUE, FALSE /* don't shorten zero */);
+
+		g_assert (timer_applet->cur_profile_name);
+		if (strlen (timer_applet->cur_profile_name) > 0) {
 			/* Resume the "<timer name>" timer? */
 			dialog_header_text = g_strdup_printf(_("Resume the \"%s\" timer?"), timer_applet->cur_profile_name);
 		}
@@ -821,7 +1041,9 @@ ask_user_resume_timer (TimerApplet *timer_applet)
 				break;
 
 			case GTK_RESPONSE_NO:
-				start_timer_dialog_open (timer_applet);
+				start_timer_dialog_open_with_name_and_duration (timer_applet,
+										timer_applet->cur_profile_name,
+										timer_applet->cur_profile_duration);
 				break;
 
 
@@ -852,14 +1074,18 @@ timer_applet_get_presets_list (void)
 /**
  * Indicate which timer preset should run. If preset_name is NULL,
  * then the timer will be a "one-time" instant timer.
+ *
+ * NOTE: preset_name must not point to the same string as
+ * timer_applet->cur_profile_name.
  */
 static void
 set_current_preset (TimerApplet *timer_applet, const gchar *preset_name, gdouble timer_duration)
 {
-	if (timer_applet->cur_profile_name) {
-		g_free (timer_applet->cur_profile_name);
-		timer_applet->cur_profile_name = NULL;
-	}
+	g_assert (preset_name != NULL);
+	g_assert (preset_name != timer_applet->cur_profile_name);
+
+	g_assert (timer_applet->cur_profile_name);
+	g_free (timer_applet->cur_profile_name);
 	timer_applet->cur_profile_name = g_strdup (preset_name);
 	timer_applet->cur_profile_duration = timer_duration;
 	set_ui_time (timer_applet, timer_duration);
@@ -867,11 +1093,14 @@ set_current_preset (TimerApplet *timer_applet, const gchar *preset_name, gdouble
 
 /**
  * Start the timer with the given preset name and duration.
- * preset_name can be NULL to indicate a one-time timer.
+ * preset_name can be an empty string to indicate a one-time timer.
+ * preset_name cannot be NULL.
  */
 void
 timer_applet_start_timer (TimerApplet *timer_applet, const gchar *preset_name, gdouble timer_duration)
 {
+	g_assert (timer_applet);
+	g_assert (preset_name);
 	g_assert (timer_applet->timer_running == FALSE);
 	g_assert (timer_duration >= 0);
 
@@ -920,7 +1149,6 @@ timer_applet_add_preset (const gchar *name, gdouble duration)
 	
 	save_timer_presets ();
 	
-	
 	return row_ref_from_iter (GTK_TREE_MODEL (timer_presets_list), &iter);
 }
 
@@ -1013,6 +1241,7 @@ timer_applet_get_remaining_time (TimerApplet *timer_applet)
 void
 timer_applet_init (TimerApplet *timer_applet, PanelApplet *panel_applet)
 {
+	BonoboUIComponent        *popup_component;
 	g_assert (timer_applet);
 	g_assert (panel_applet);
 	
@@ -1025,6 +1254,18 @@ timer_applet_init (TimerApplet *timer_applet, PanelApplet *panel_applet)
 		/* Perform one-time initialization */
 
 		timer_presets_list = gtk_list_store_new (3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_DOUBLE);
+		
+		presets_changed_sig_id = g_signal_newv ("presets-changed",
+						G_TYPE_FROM_CLASS (GTK_LIST_STORE_GET_CLASS (timer_presets_list)),
+						G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
+						NULL /* class closure */,
+						NULL /* accumulator */,
+						NULL /* accu_data */,
+						g_cclosure_marshal_VOID__VOID,
+						G_TYPE_NONE /* return type */,
+						0 /* n_params */,
+						NULL /* param_types */);
+						
 		load_timer_presets ();
 
 		gnome_window_icon_set_default_from_file (PIXMAPS_DIR "/timer-applet-icon.png");
@@ -1035,7 +1276,7 @@ timer_applet_init (TimerApplet *timer_applet, PanelApplet *panel_applet)
 		
 		about_dialog = NULL;
 	}
-	timer_applet->cur_profile_name = NULL;
+	timer_applet->cur_profile_name = g_strdup ("");
 	timer_applet->cur_profile_duration = 0;
 	
 	timer_applet->main_button = NULL;
@@ -1070,8 +1311,6 @@ timer_applet_init (TimerApplet *timer_applet, PanelApplet *panel_applet)
 	timer_applet->prefs_dialog = prefs_dialog_new (timer_applet->applet);
 	g_assert (timer_applet->prefs_dialog);
 	
-	timer_applet->notification_sound_path = SOUNDS_DIR "/clicked.wav";
-	
 	construct_applet_content (timer_applet);
 	
 	/* Setup right-click menu */
@@ -1081,8 +1320,28 @@ timer_applet_init (TimerApplet *timer_applet, PanelApplet *panel_applet)
 					   NULL,
 					   timer_applet_verbs,
 					   timer_applet);
-	
+	popup_component = panel_applet_get_popup_component (PANEL_APPLET (timer_applet->applet));
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_PAUSE_TIMER,
+					"hidden", "1",
+					NULL);
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_RESUME_TIMER,
+					"hidden", "1",
+					NULL);
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_RESTART_TIMER,
+					"hidden", "1",
+					NULL);
+	bonobo_ui_component_set_prop (popup_component,
+					TA_MENU_COMMAND_STOP_TIMER,
+					"hidden", "1",
+					NULL);
+
 	start_timer_dialog_create (timer_applet);
+	update_presets_submenu (timer_applet);
+	
+	timer_applet->presets_changed_handler_id = g_signal_connect (G_OBJECT (timer_presets_list), "presets-changed", (GCallback)on_presets_changed, timer_applet);
 	
 	g_timeout_add (TIMER_TIMEOUT_INTERVAL, (GSourceFunc) on_timer_timeout, timer_applet);
 
@@ -1108,6 +1367,87 @@ timer_applet_init (TimerApplet *timer_applet, PanelApplet *panel_applet)
 /**
  * Callbacks for context menu
  */
+
+static void
+on_pause_timer_activate (BonoboUIComponent *component, TimerApplet *timer_applet, const gchar *cname)
+{
+	g_assert (timer_applet);
+
+	pause_timer (timer_applet);
+}
+
+static void
+on_resume_timer_activate (BonoboUIComponent *component, TimerApplet *timer_applet, const gchar *cname)
+{
+	g_assert (timer_applet);
+
+	resume_timer (timer_applet);
+}
+
+static void
+on_restart_timer_activate (BonoboUIComponent *component, TimerApplet *timer_applet, const gchar *cname)
+{
+	gchar *preset_name;
+	g_assert (timer_applet);
+
+	stop_timer (timer_applet);
+
+	g_assert (timer_applet->cur_profile_duration >= 0);
+
+	/* We need to create a copy of the string first because cur_profile_name will be
+	 * freed before being set to the new preset name (which in this case is the same name).
+	 */
+	preset_name = g_strdup (timer_applet->cur_profile_name);
+	timer_applet_start_timer (timer_applet, preset_name, timer_applet->cur_profile_duration);
+	g_free (preset_name);
+	preset_name = NULL;
+}
+
+static void
+on_stop_timer_activate (BonoboUIComponent *component, TimerApplet *timer_applet, const gchar *cname)
+{
+	g_assert (timer_applet);
+
+	stop_timer (timer_applet);
+}
+	
+static void
+on_presets_submenu_item_activate (BonoboUIComponent *uic, TimerApplet *timer_applet, const gchar *verb)
+{
+	int preset_index;
+	
+	g_assert (timer_applet);
+	
+	if (sscanf (verb, "Preset_%d", &preset_index) != 1) {
+		g_print ("ERROR: Invalid preset submenu verb: %s\n", verb);
+		return;
+	}
+
+	GtkTreeIter iter;
+	{
+		GtkTreePath *path = gtk_tree_path_new_from_indices (preset_index, -1);
+		gboolean iter_set = gtk_tree_model_get_iter (GTK_TREE_MODEL (timer_presets_list), &iter, path);
+		g_assert (iter_set);
+		gtk_tree_path_free (path);
+	}
+
+	{
+		gchar *preset_name;
+		gdouble preset_duration;
+				
+		gtk_tree_model_get (GTK_TREE_MODEL (timer_presets_list), &iter,
+							PROFILE_NAME_COL, &preset_name,
+							PROFILE_DURATION_COL, &preset_duration,
+							-1);
+		if (timer_is_running (timer_applet) || timer_is_paused (timer_applet)) {
+			stop_timer (timer_applet);
+		}
+		timer_applet_start_timer (timer_applet, preset_name, preset_duration);
+		g_print ("Activated this preset: %s (index = %d)\n", preset_name, preset_index);
+		g_free (preset_name);
+		preset_name = NULL;
+	}
+}
  
 static void
 on_manage_presets_activate (BonoboUIComponent *component, TimerApplet *timer_applet, const gchar *cname)
@@ -1151,7 +1491,6 @@ on_about_activate (BonoboUIComponent *component, TimerApplet *timer_applet, cons
 	about_dialog_show (&about_dialog);
 }
 
-
 /**
  * Other callbacks
  */
@@ -1173,7 +1512,6 @@ on_main_button_clicked (GtkWidget *button, TimerApplet *timer_applet)
 	}
 }
 
-
 /**
  * A hack used in construct_applet_content to allow middle-clicks and right-clicks in the applet
  */
@@ -1197,17 +1535,24 @@ on_button_press_hack (GtkWidget *widget, GdkEventButton *event, TimerApplet *tim
 static void
 on_preferences_change (GConfClient *client, guint cnxn_id, GConfEntry *entry, TimerApplet *timer_applet)
 {
-	prefs_dialog_update (GTK_DIALOG (timer_applet->prefs_dialog));
 	if (timer_applet->applet) {
-		update_main_button (timer_applet);
-	}
-	else {
-		GConfClient *gconf_client;
+		prefs_dialog_update (GTK_DIALOG (timer_applet->prefs_dialog));
 
-		gconf_client = gconf_client_get_default ();
-		gconf_client_notify_remove (gconf_client, cnxn_id);
-		
-		g_object_unref (G_OBJECT (gconf_client));
+		/* If this handler is called while the applet is being removed from the panel,
+		 * prefs_dialog_update() can cause the applet's "destroy" handler to be run
+		 * before the code below gets to execute.
+		 * Specifically, the "destroy" handler is run when prefs_dialog_update() changes
+		 * the "active" property of one of its checkboxes from TRUE to FALSE or FALSE to
+		 * TRUE. The "destroy" handler, however, does NOT run if a checkbox's value is
+		 * unchanged, such as setting it to TRUE when it's already TRUE or setting it to
+		 * FALSE when it's already FALSE. 
+		 *
+		 * Therefore, we check again to see if timer_applet->applet is NULL, which indicates
+		 * that the applet's "destroy" handler has already been run.
+		 */
+		if (timer_applet->applet) {
+			update_main_button (timer_applet);
+		}
 	}
 }
 
@@ -1244,6 +1589,15 @@ on_timer_timeout (TimerApplet *timer_applet)
 	}
 }
 
+static void
+on_presets_changed (GObject *obj, TimerApplet *timer_applet)
+{
+	g_assert (timer_applet);
+	
+	g_print ("Presets changed.\n");
+	update_presets_submenu (timer_applet);
+}
+
 static void
 on_applet_change_background (PanelApplet *applet, PanelAppletBackgroundType type,
 				GdkColor *color, GdkPixmap *pixmap, gpointer user_data)
@@ -1305,10 +1659,11 @@ on_applet_instance_destroy (GtkWidget *applet, TimerApplet *timer_applet)
 		gconf_client = gconf_client_get_default ();
 		
 		gconf_client_notify_remove (gconf_client, timer_applet->gconf_notification_id);
-		
 		timer_applet->gconf_notification_id = 0;
 		
 		g_object_unref (G_OBJECT (gconf_client));
 	}
-}
 
+	g_signal_handler_disconnect (G_OBJECT (timer_presets_list), timer_applet->presets_changed_handler_id);
+	timer_applet->presets_changed_handler_id = 0;
+}
