@@ -21,9 +21,10 @@
 #include <gtk/gtk.h>
 #include <panel-applet.h>
 
-#define TIMER_APPLET_GCONF_PATH			"/apps/timer-applet"
 #define SHOW_REMAINING_TIME_KEY			"show_remaining_time"
 #define PLAY_NOTIFICATION_SOUND_KEY		"play_notification_sound"
+#define USE_CUST_NOTIFICATION_SOUND_KEY		"use_custom_notification_sound"
+#define CUST_NOTIFICATION_SOUND_PATH_KEY	"custom_notification_sound_path"
 #define NOTIFICATION_DIALOG_STICKY_KEY		"notification_dialog_sticky"
 #define NOTIFICATION_DIALOG_ALWAYS_ON_TOP_KEY	"notification_dialog_always_on_top"
 #define MAX_NUM_HOURS				48
@@ -75,9 +76,9 @@ typedef struct _TimerApplet
 	/* Preferences dialog */
 	GtkWidget *prefs_dialog;
 	
-	gchar *notification_sound_path;
-	
+
 	guint gconf_notification_id;
+	gulong presets_changed_handler_id;
 } TimerApplet;
 
 void
